#include <cstdarg>
#include <vector>
#include <algorithm>

#include "level.hh"
#include "objects.hh"

#include "treegen.hh"
#include "legofill.hh"

/* Wall details are called greebles/nurnies. */

static void GenericAsciiMaker(Screen& s, int n,...)
{
    char map[0x100*4];
    for(unsigned a=0; a<sizeof(map); ++a) map[a] = '?';
    va_list ap;
    va_start(ap,n);
    char ch = '?';
    while(n != -2)
    {
        if(n==-1)
            ch = va_arg(ap, int);
        else
        {
            int p = va_arg(ap, int)&3;
            map[n*4+p] = ch;
        }
        n = va_arg(ap, int);
    }
    va_end(ap);
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = s.outmap[y*16+x].tile;
            unsigned char p = s.outmap[y*16+x].palette;
            s.ascii[y*32+   x] = map[c*4+p];
            s.ascii[y*32+16+x] = ' ';
        }
}

void DummyAscii(Screen& s)
{
    for(unsigned y=0; y<16; ++y)
    {
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = s.outmap[y*16+x].tile;
            char ch = c;
            s.ascii[y*32+x] = ch;
            s.ascii[y*32+x+16] = ' ';
        }
    }
}
void DummyChar(Screen& s)
{
    // Convert to tiles
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c   = 0;
            unsigned char pal = 0;
            
            c   = s.ascii[y*32+x];
            pal = 0x01;
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = pal;
        }
}
/////////////////////////////////////////////////////////////

void CutmanAscii(Screen& s)
{
    GenericAsciiMaker(s,
// Sky
-1,'.', 0x01,1, // cyan background
-1,'V', 0x00,0, // black background
// Background
-1,'W', 0x23,3, // window top
        0x24,3, // window bottom
-1,' ', 0x26,2, 0x25,2, // background, shadow
        0x2C,2, 0x2B,2, // inset decoration, shadowed
        0x34,2, 0x35,2, 0x3C,2, // exposed pipes (begin,repeat,end)
// Walls
-1,'#', 0x62,0, // step
-1,'#', 0x68,0, 0x69,0, // walls
-1,'#', 0x51,0, 0x52,0, 0x53,0, // Wily signature
// Wily door
-1,'D', 0x10,0,
// Gutsblock
-1,'G', 0x38,0, 0x39,0,
        0x3A,0, 0x3B,0,
// Ladder
-1,'|', 0x82,0,
// Spike
-1,'^', 0xC4,1, // sky bg
// Scissor machine
-1,'C', 0x20,2, 0x21,2, 0x2A,2, // top side
        0x3D,2, // left edge
        0x2D,2, // ladders
        0x2E,2, 0x2F,2, // pipes
        0x27,2, // right edge
        0x36,2, 0x37,2,
        0x3E,2, 0x3F,2, // scissor hole
-2);
}

void CutmanChar(Screen& s)
{
    std::srand(0);
    
    char decorations[16*16];
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            char c;
            switch(std::rand() % 6)
            {
                case 0: c = 1; break;
                case 1: c = 2; break;
                default: c = 0; break;
            }
            decorations[y*16+x] = c;
        }


    // Convert to tiles
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c   = 0;
            unsigned char pal = 0;
            
            /* Cutman palettes:
             *   
             * 0 = solid
             * 1 = blacks
             * 2 = blues
             * 3 = blue,gray
             */
            
            switch(s.ascii[y*32+x])
            {
                case '.':
                    c   = 0x01; // background
                    pal = 0x01;
                    break;
                case 'V':
                    c   = 0x00; // background
                    pal = 0x00;
                    break;
                case 'C': // ignore scissor machine...
                case ' ': // background walls
                    if(y > 0 &&
                        (s.ascii[(y-1)*32+x] != '.'
                      && s.ascii[(y-1)*32+x] != ' ')
                      )
                        c = 0x25; // shadow
                    else
                    {
                        switch(decorations[y*16+x])
                        {
                            case 1:
                                c = 0x2C; // inset decoration
                                break;
                            case 2:
                                c = 0x35;
                                if(x<16
                                && decorations[y*16+x+1]==2
                                && (x==0 || decorations[y*16+x-1]!=2))
                                    c = 0x34; // beginning

                                if(x>0  && s.ascii[y*32+x-1] == ' '
                                && x<15 && s.ascii[y*32+x+1] == ' ')
                                {
                                    break;
                                }
                                // passthru
                            case 0:
                                c = 0x26; // no decoration
                        }
                    }
                    pal = 0x02;
                    break;
                case 'W': // background window
                {
                    c = 0x23;
                    if(y > 0 && s.ascii[(y-1)*32+x] == 'W') c = 0x24;
                    pal = 0x02;
                    break;
                }
                case '#':
                case 'R':
                {
                    bool step_here =
                        y > 0 && (s.ascii[(y-1)*32+x] != s.ascii[y*32+x]);
                    bool step_left = false;/*
                        x > 0 && s.ascii[y*32+x-1] == s.ascii[y*32+x] &&
                        y > 0 && (s.ascii[(y-1)*32+x-1] != s.ascii[y*32+x-1]);*/
                    bool step_right = false;/*
                        x < 15 && s.ascii[y*32+x+1] == s.ascii[y*32+x] &&
                        y > 0 && (s.ascii[(y-1)*32+x+1] != s.ascii[y*32+x+1]);*/
                    
                    if(step_here || step_left || step_right)
                        c = 0x62; // step
                    else
                        c = 0x68 + ((x+y) & 1);
                    pal = 0x00;
                    if(s.ascii[y*32+x]=='R') pal=3;
                    break;
                }
                case Ascii_Ladder:
                    c   = 0x82; // ladder
                    pal = 0x00;
                    if((x>0 && s.ascii[y*32+x-1]=='R')
                    || (x<15&& s.ascii[y*32+x+1]=='R'))
                    {
                        pal = 0x03;
                        if((x>0 && s.ascii[y*32+x-1]=='#')
                        || (x<15&& s.ascii[y*32+x+1]=='#'))
                            pal = 0x00;
                    }
                    break;
                case Ascii_Spike:
                    c   = 0xC4; // spike
                    if(y>0 && s.ascii[(y-1)*32+x] == ' ')
                        pal = 0x03;
                    else
                        pal = 0x01;
                    break;
                case Ascii_Guts:
                    c   = 0x38 + (x&1) + (y&1)*2; // gutsblock
                    pal = 0x00;
                    break;
                case Ascii_WilyDoor:
                    c   = 0x10;
                    pal = 0x00;
                    break;
                default:
                    printf("? '%c'\n", s.ascii[y*32+x]);
            }
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = pal;
        }
}

/////////////////////////////////////////////////////////////

/* ICEMAN

ti pa

30.1 = 1st row cloud

32.1
01.1 = 2-3 row background

31.1 {30.1} 31.1
33.1 {34.1} 35.1 = 2-3 row cloud

01.1 = other row background

33.1 {34.1} 35.1 = other row cloud

3F.2 = frozen grass

3E.2 = frozen tree branch

38.2 39.2 3A.2 
3B.2 3C.2 3D.2 = frozen tree leaves

A0.0 A1.0 = pillar surface
68.0 69.0 = pillar

         66.0 67.0
         6E.0 6F.0
77.0 ... 76.0 6F.0 = metal object

A2.0 = vertical metal bar surface
6A.0 = vertical metal bar
6B.0 = vertical metal bar detail

E3.3 = water surface
C1.3 = water

2D.3 = inside
2C.3 = inside shadow

24.0 25.0 = plastic pit (background)

*/

void IcemanAscii(Screen& s)
{
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = s.outmap[y*16+x].tile;
            unsigned char p = s.outmap[y*16+x].palette;
            
            char ch = '?';
            /* Classify types.
             *
             *
             *
             */
            switch(c)
            {
                // sky:
                case 0x32:
                case 0x01:
                case 0x31: case 0x30:
                case 0x33: case 0x34: case 0x35: // clouds, sky.
                case 0x00: //void
                case 0x3E: case 0x3F:
                case 0x38 ... 0x3A:
                case 0x3B ... 0x3D: // trees and grass
                    ch = '.';
                    break;
                
                case 0xA0: case 0xA1:
                case 0x68: case 0x69: // pillar
                case 0xA2: case 0x6A: case 0x6B: // metal part in pillar
                    ch = '#'; // pillar
                    break;
                case 0x51: case 0x52: case 0x53: // wily signature
                    ch = '#';
                    break;
                case 0x77: // horizontal metal bar (part of "metal object")
                    ch = '=';
                    break;
                case 0x66: case 0x67:
                case 0x6E: case 0x6F:
                case 0x76: // metal object
                    ch = 'M';
                    break;
                case 0x2C: case 0x2D:
                    ch = ' '; // inside
                    break;
                case 0xC1: case 0xE3:
                    ch = '~'; // water
                    break;
                case 0x24: case 0x25:
                    ch = 'V';
                    break;
                    
                case 0x10:
                    ch = Ascii_WilyDoor;
                    break;
                
                // There are no ladders, gutsblock or spikes in Iceman level.
            }
            
            s.ascii[y*32+x] = ch;
            s.ascii[y*32+x+16] = ' ';
            //printf("%02X",c);      
        }
}

namespace IceUtil
{
    bool isFree(unsigned Wid, unsigned Hei, unsigned BranchWid,
                unsigned x, unsigned y, unsigned BranchHei,
                const unsigned char* screen,
                const char* ground)
    {
        unsigned mask = 0;
		for(unsigned i = 0; i < Wid; ++i)
			for(unsigned j = 0; j < Hei; ++j)
			{
				if(ground[x+i + 32*(y+j)] != '.') return false;
				if(screen[x+i + 16*(y+j)] != 0) ++mask;
	        }
	    if(mask != 0 && mask != Wid*Hei) return false;

		const unsigned bx = x + Wid/2 - BranchWid/2;
		const unsigned by = y + Hei;

		for(unsigned iy=0; iy<BranchHei; ++iy)
			for(unsigned ix = 0; ix < BranchWid; ++ix)
				if(ground[bx+ix + 32*(by+iy)] != '.') return false;
		return true;
    }

    void CreateTrees(unsigned char dest[16*16], const char ground[32*16])
	{
		static const unsigned char SmallBall[2][3] =
		{
			{ 0x38, 0x39, 0x3A },
			{ 0x3B, 0x3C, 0x3D }
		};
		static const unsigned char SmallBranch[1] = { 0x3E };
		
		for(unsigned y=10; y<16; ++y)
		    for(unsigned tryn=0; tryn<40; ++tryn)
		    {
		        unsigned x = rand()%14 + 1;
		        char h = ground[y*32+x];
		        bool here = (h=='#' || h=='M' || h=='=');
		        bool above= ground[(y-1)*32+x] == '.';
		        if(here && above)
		        {
		            unsigned hei = std::rand()%2 + 1;
		            
		            unsigned beginY = y-2-hei;
		            if(isFree(3,2,1, x,beginY,hei, dest,ground))
		                CreateTree<3,2,1> (SmallBall,SmallBranch, dest, x,beginY, hei);
		            else
		                dest[(y-1)*16+x] = 0x3F;
		        }
		    }
	}
}


void IcemanChar(Screen& s)
{
    unsigned randseed = 0;
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            randseed += s.ascii[y*32+x];
    std::srand(randseed);
    
    unsigned char decorations[16*16] = {0};
    
    for(unsigned y=2; y<6; ++y)
    {
        for(unsigned x=0; x<16; )
        {
            unsigned nextcloud = 0 + rand()%(y*2);
            x += nextcloud;
            if(x >= 14) break;
            unsigned width = 2 + rand()%(1+rand()%3);
            if(x+width > 16) width=16-x;
            for(unsigned xm=0; xm<width; ++xm)
            {
                char c = 2;
                if(xm==0) c=1; else if(xm==width-1) c=3;
                decorations[y*16+x+xm] = c;
            }
            x += width;
        }
    }
    
    IceUtil::CreateTrees(decorations, s.ascii);

    for(unsigned y=1; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            if(s.ascii[y*32+x] == '#'
            && s.ascii[(y-1)*32+x] != '#'
            && rand()%6 == 0)
            {
                for(unsigned my=0; y+my<16 && s.ascii[(y+my)*32+x]=='#'; ++my)
                {
                    decorations[(y+my)*16+x] = 1;
                    if(my>0 && rand()%4 == 0)break;
                }
            }
    
    
    // Convert to tiles
    for(unsigned y=0; y<16; ++y) 
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c   = 0;
            unsigned char pal = 0;
            
            switch(s.ascii[y*32+x])
            {
                case '.':
                {
                    // background
                    pal = 1;
                    
                    switch(y)
                    {
                        case 0:
                            c = 0x30; break;
                        case 1:
                            c = 0x32; if(decorations[2*16+x]) c = decorations[2*16+x]==2 ? 0x30 : 0x31; break;
                        case 2:
                            c = 0x01; if(decorations[2*16+x]) c = 0x33+decorations[2*16+x]-1; break;
                        default:
                        {
                            char deco = decorations[y*16+x];
                            c = 0x01;
                            if(deco>=1 && deco<=3) c = 0x33+decorations[y*16+x]-1;
                            else if(deco>3) { c=deco; pal=2; }
                            break;
                        }
                    }
                    break;
                }
                case '#':
                {
                    pal = 0;
                    bool decor = decorations[y*16+x] == 1;
                    bool up = y>0 && s.ascii[(y-1)*32+x]!='#';
                    if(up) c = decor ? 0xA2 : (0xA0 + (x&1));
                    else c = decor ? (0x6A + !(rand()%4))
                                   : (0x68 + (x&1));
                    break;
                }
                case '=':
                    c = 0x77;
                    pal = 0;
                    break;
                case 'M':
                {
                    pal = 0;
                    bool left  = x>0  && s.ascii[y*32+(x-1)]=='M';
                    bool up    = y>0  && s.ascii[(y-1)*32+x]=='M';
                    bool down  = y<15 && s.ascii[(y+1)*32+x]=='M';
                    if(!up)
                        c = 0x66 + left;
                    else if(down)
                        c = 0x6E + left;
                    else
                        c = left ? 0x6F : 0x76;
                    break;
                }
                case 'V':
                    pal = 0;
                    c = 0x24 + (x&1);
                    break;
                case '~':
                {
                    bool up = y>0 && s.ascii[(y-1)*32+x]=='~';
                    pal = 3;
                    c = up ? 0xC1 : 0xE3;
                    break;
                }
                case ' ':
                {
                    bool up  = y==0 || s.ascii[(y-1)*32+x]==' ';
                    c   = up ? 0x2D : 0x2C;
                    pal = 3;
                    // inside
                    break;
                }
                case Ascii_Ladder:
                    c   = 0x82; // ladder
                    pal = 0x00;
                    break;
                case Ascii_Spike:
                    c   = 0xC4; // spike
                    pal = 0x01;
                    break;
                case Ascii_WilyDoor:
                    c   = 0x10;
                    pal = 0x00;
                    break;
                default:
                    printf("? '%c'\n", s.ascii[y*32+x]);
            }
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = pal;
        }
}

/////////////////////////////////////////////////////////////

/* BOMBMAN

ti pa
00.0 = void
01.3 = sky

24.3 25.3 26.3 27.3
2C.3 33.3 33.3 2F.3
34.3 33.3 33.3 37.3
3C.3 3D.3 3E.3 3F.3 = big ball

38.3 39.3 = big branch

2D.3 2E.3
35.3 36.3 = small ball

3A.3 3B.3 = small branch

61.0 = horizontal pipe
60.1 = pipe left edge
62.1 = pipe right edge

2B.2 = inside background
23.2 = inside shadow

68.0 {69.0} 6A.0
70.0 {71.0} 72.0 = pillar

83.0 = ladder

C4.1 = spike up

51.0 52.0 53.0 = wily ornament
10.0 = wily door



*/
void BombmanAscii(Screen& s)
{
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = s.outmap[y*16+x].tile;
            unsigned char p = s.outmap[y*16+x].palette;
            
            char ch = '?';
            /* Classify types.
             *
             *
             *
             */
            switch(c)
            {
                // sky:
                case 0x01: // sky
                case 0x24 ... 0x27:
                case 0x2C: case 0x33: case 0x2F:
                case 0x34: case 0x37:
                case 0x3C ... 0x3F: // big tree
                case 0x38 ... 0x39: // big branch
                case 0x2D ... 0x2E:
                case 0x35 ... 0x36: // small tree
                case 0x3A ... 0x3B: // small branch
                case 0x00: // void
                    ch = '.';
                    break;
                
                case 0x60 ... 0x62:
                    ch = '=';
                    break;
                case 0x68 ... 0x6A:
                case 0x70 ... 0x72:
                case 0x51 ... 0x53:
                    ch = '#';
                    break;
                case 0x23: case 0x2B:
                    ch = ' '; // inside
                    break;
                case 0x83:
                    ch = Ascii_Ladder;
                    break;
                case 0xC4:
                    ch = Ascii_Spike;
                    break;
                    
                case 0x10:
                    ch = Ascii_WilyDoor;
                    break;
                
                // There are no gutsblocks in Bombman level.
            }
            
            s.ascii[y*32+x] = ch;
            s.ascii[y*32+x+16] = ' ';
            //printf("%02X",c);      
        }
}

namespace BombUtil
{
    bool isFree(unsigned size, unsigned x, unsigned y,
                const unsigned char* screen,
                const char* ground)
    {
        bool hitsGround = false;

        unsigned FreeSpace=0, HitSpace=0;
        for(unsigned iy = y; iy < y+size; ++iy)
            for(unsigned ix = x; ix < x+size; ++ix)
            {
                if(screen[ix+16*iy] != 0) return false;
                if(ground[ix+32*iy] != '.') ++HitSpace; else ++FreeSpace;
            }
        if(HitSpace > FreeSpace) return false;

        const unsigned bx = x+size/2-1;
        for(unsigned iy = y+size-1; iy < 16; ++iy)
        {
            unsigned groundHits = 0;
            for(unsigned ix = bx; ix < bx+2; ++ix)
            {
                if(screen[ix+16*iy] != 0) return false;
                if(ground[ix+32*iy] != '.') ++groundHits;
                if(ground[ix+32*iy] == '^') return false;
            }
            if(groundHits == 2) hitsGround = true;
        }

        return hitsGround;
    }

    void CreateTrees(unsigned char dest[16*16], const char ground[32*16])
	{
        const unsigned MaxTrees = 8, MaxTries = 100;
        const unsigned LowerSideOffset = 2;

		static const unsigned char BigBall[4][4] =
		{
			{ 0x24, 0x25, 0x26, 0x27 },
			{ 0x2C, 0x33, 0x33, 0x2F },
			{ 0x34, 0x33, 0x33, 0x37 },
			{ 0x3C, 0x3D, 0x3E, 0x3F }
		};
		static const unsigned char BigBranch[2] = { 0x38, 0x39 };
		static const unsigned char SmallBall[2][2] =
		{
			{ 0x2D, 0x2E },
			{ 0x35, 0x36 }
		};
		static const unsigned char SmallBranch[2] = { 0x3A, 0x3B };

		unsigned tries = 0;
		for(unsigned treeCount = 0; treeCount < MaxTrees; ++treeCount)
		{
			while(true)
			{
				if(++tries > MaxTries) return;
	 
				bool big = std::rand() % 3; // Big trees more probable than small ones
				if(big)
				{
					unsigned x = std::rand() % (16 - 4);
					unsigned y = std::rand() % (16 - 4 - 2);
					if(!isFree(sizeof(BigBall[0]), x, y, dest, ground)) continue;
					CreateTree<4,4,2> (BigBall, BigBranch, dest, x, y,
					           16-y-4);
					break;
				}
				else
				{
					unsigned x = std::rand() % (16 - 2);
					unsigned y = std::rand() % (16 - 2 - 2);
					if(!isFree(sizeof(SmallBall[0]), x, y, dest, ground)) continue;
					CreateTree<2,2,2> (SmallBall, SmallBranch, dest, x, y,
					           16-y-2);
					break;
				}
			}
		}
	}
}

void BombmanChar(Screen& s)
{
    unsigned randseed = 0;
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            randseed += s.ascii[y*32+x];
    std::srand(randseed);
    
    unsigned char decorations[16*16] = { 0 };
    BombUtil::CreateTrees(decorations, s.ascii);
    
    // Convert to tiles
    for(unsigned y=0; y<16; ++y) 
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c   = 0;
            unsigned char pal = 0;
            
            switch(s.ascii[y*32+x])
            {
                case '.':
                {
                    // background
                    char deco = decorations[y*16+x];
                    c   = deco ? deco : 0x01;
                    pal = 3;
                    break;
                }
                case '#':
                {
                    bool left  = x>0  && s.ascii[y*32+(x-1)]=='#';
                    bool right = x<15 && s.ascii[y*32+(x+1)]=='#';
                    bool up    = y==0 || s.ascii[(y-1)*32+x]!='.';
                    // walls
                    c   = 0x71;
                    if(!up) c -= 8;
                    if(!left && right) c -= 1;
                    else if(!right &&left) c += 1;
                    pal = 0;
                    break;
                }
                case ' ':
                {
                    bool up  = y==0 || s.ascii[(y-1)*32+x]==' ';
                    c   = up ? 0x2B : 0x23;
                    pal = 2;
                    // inside
                    break;
                }
                case '=':
                {
                    bool left  = x>0  && s.ascii[y*32+(x-1)]!='.';
                    bool right = x<15 && s.ascii[y*32+(x+1)]!='.';
                    // pipes
                    c   = 0x61;
                    pal = 0;
                    if(!left && right)     { c -= 1; pal=1; }
                    else if(!right &&left) { c += 1; pal=1; }
                    break;
                }
                case Ascii_Ladder:
                    c   = 0x82; // ladder
                    pal = 0x00;
                    break;
                case Ascii_Spike:
                    c   = 0xC4; // spike
                    pal = 0x01;
                    break;
                case Ascii_WilyDoor:
                    c   = 0x10;
                    pal = 0x00;
                    break;
                default:
                    printf("? '%c'\n", s.ascii[y*32+x]);
            }
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = pal;
        }
}

/////////////////////////////////////////////////////////////

void FiremanAscii(Screen& s)
{
    GenericAsciiMaker(s,
// Sky
-1,'.', 0x00,0, // black background
// Background
-1,' ', 0x28,2, // red background
        0x29,2, // red background - shadow under a platform
-1,' ', 0x22,2, // window in red background
        0x23,2, // window in red background - shadow under a platform
        0x2B,2, // hole in a window
-1,'-', 0x27,2, // red thin horizontal pipe on black background
        0x26,2, // small detail in red thin pipe
        0x2F,2, // big detail in red thin pipe
-1,'=', 0x27,1, // green thin horizontal pipe on black background
        0x26,1, // small detail in green thin pipe
        0x2F,1, // big detail in green thin pipe
// Green background
-1,'b', 0x28,1, // green background
        0x29,1, // green background - shadow under a platform
// Lava (background)
-1,'~', 0x30,3, // lava surface
-1,'L', 0x38,3, // lava
-1,'z', 0xE4,0, // fence
// Lava (foreground)
-1,'E', 0x73,0,
        0x7B,0,         // horizontal lava tunnel 1x2, (stack these horizontally)
-1,'W', 0x6C,0, 0x6D,0, // vertical lava tunnel 2x1, (stack these vertically)
-1,'%', 0x74,0, 0x75,0,
        0x7C,0, 0x7D,0, // lava flow junction (2x2)
-1,'/', /*0,3,*/0x36,3,
        0x3F,3, 0x3E,3, // lava flow corner "/"
// Torch
-1,'>', 0x65,0, // right
-1,'^', 0x6E,0, // up
// Walls
-1,'#', 0x61,0, // wall
-1,'#', 0x60,0, // walkable surface
-1,'#', 0x77,0, // wall detail (best used in horizontal stretches)
-1,'#', 0x51,0, 0x52,0, 0x53,0, // Wily signature
// Red walls
-1,'&', 0x61,2, // wall
-1,'&', 0x60,2, // walkable surface
-1,'&', 0x77,2, // wall detail (best used in horizontal stretches)
// Wily door
-1,'D', 0x10,0,
// Ladders
-1,'|', 0x83,0,
-1,']', 0x82,0,

-2);
}

void FiremanChar(Screen& s)
{
    unsigned randseed = 0;
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            randseed += s.ascii[y*32+x];
    std::srand(randseed);
    
    char decorations[16*16] = { 0 };
    
    // Put windows somewhere
    for(unsigned y=0; y<16; ++y) 
        for(unsigned x=0; x<16; ++x)
            if(s.ascii[y*32+x]==' ' || s.ascii[y*32+x]=='b')
                if(rand()%30 == 0)
                    decorations[y*16+x] = 1;
    
    // Extend those windows randomly
    for(unsigned n=0; n<8; ++n)
    {
		for(unsigned y=1; y<15; ++y) 
			for(unsigned x=1; x<15; ++x)
			  if(rand()%3 == 0)
			    if(decorations[(y-1)*16+(x)]==1
			    || decorations[(y+1)*16+(x)]==1
			    || decorations[(y)*16+(x-1)]==1
			    || decorations[(y)*16+(x+1)]==1)
			        decorations[y*16+x] = 1;
    }
    

    // Put streaks in some walls
    for(unsigned n=0; n<5; ++n)
    {
        unsigned y = rand()%16;
        unsigned x1 = rand()%13;
        unsigned x2 = x1 + 2 + rand()%6;
        if(x2 > 15) x2 = 15;
        for(unsigned x=x1; x<=x2; ++x)
            if(decorations[y*16+x] == 0)
                decorations[y*16+x] = 2;
    }

    // Convert to tiles
    for(unsigned y=0; y<16; ++y) 
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c   = 0;
            unsigned char pal = 0;
            
            switch(s.ascii[y*32+x])
            {
                case '.':
                {
                    c   = 0x00; // black background
                    pal = 0x00;
                    break;
                }
                case '-':
                case '=':
                {
					pal = s.ascii[y*32+x]=='-' ? 0x02 : 0x01;
					c   = 0x27;
					switch(rand()%10)
					{
						case 0: c = 0x2F; break;
						case 1: c = 0x26; break;
					}                        
                    break;
                }
                case '>': c=0x65; pal=0; break;
                case '^': c=0x6E; pal=0; break;
                case ' ':
                case 'b':
                    if(y > 0 &&
                        (s.ascii[(y-1)*32+x] != '.'
                      && s.ascii[(y-1)*32+x] != ' '
                      && s.ascii[(y-1)*32+x] != '-'
                      && s.ascii[(y-1)*32+x] != 'b')
                      )
                    {
                        c = 0x29; // shadow
                        if(decorations[y*32+x] == 1)
                            c = 0x23;
                    }
                    else
                    {
                        c = 0x28;
                        if(decorations[y*32+x] == 1)
                        {
                            c = 0x22;
                            if(rand()%5 == 0) c = 0x2B;
                        }
                    }
                    pal = s.ascii[y*32+x]==' ' ? 0x02 : 0x01;
                    break;
                case 'E':
                {
                    unsigned ypos=0;
                    while(y > ypos && s.ascii[(y-ypos-1)*32+x]=='E') ++ypos;
                    c   = (ypos&1) ? 0x7B : 0x73;
                    pal = 0x00;
                    break;
                }
                case 'W':
                {
                    unsigned xpos=0;
                    while(x > xpos && s.ascii[y*32+(x-xpos-1)]=='W') ++xpos;
                    c   = (xpos&1) ? 0x6D : 0x6C;
                    pal = 0x00;
                    break;
                }
                case '%':
                {
                    unsigned xpos=0;
                    while(x > xpos && s.ascii[y*32+(x-xpos-1)]=='%') ++xpos;
                    unsigned ypos=0;
                    while(y > ypos && s.ascii[(y-ypos-1)*32+x]=='%') ++ypos;
                    c   = 0x74 + (xpos&1) + 8*(ypos&1);
                    pal = 0x00;
                    break;
                }
                case '#':
                case '&':
                    if(y > 0 &&
                        (s.ascii[(y-1)*32+x] != '#'
                      && s.ascii[(y-1)*32+x] != '&')
                      )
                        c = 0x60; // step
                    else
                        c = decorations[y*16+x] == 2 ? 0x77 : 0x61;
                    pal = s.ascii[y*32+x]=='#' ? 0 : 2;
                    break;
                case '|':
                case ']':
                    c   = 0x82 + rand()%2; // ladder
                    pal = 0x00;
                    break;
                case Ascii_WilyDoor:
                    c   = 0x10;
                    pal = 0x00;
                    break;
                case '~':
                    c = 0x30; pal = 0x03; break;
                case 'L':
                    c = 0x38; pal = 0x03; break;
                case 'z':
                    c = 0xE4; pal = 0x00; break;
                default:
                    printf("? '%c'\n", s.ascii[y*32+x]);
            }
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = pal;
        }
}

/////////////////////////////////////////////////////////////

void ElecmanAscii(Screen& s)
{
    GenericAsciiMaker(s,
// Sky
-1,'.', 0x01,2,
// Void
-1,',', 0x00,0,
// Background
-1,' ', 0x24,3, // branch: wire1 up, wire2 left, wire3 right
        0x25,3, // horizontal wire3
        0x26,3, // branch: wire3 left, wire2 right, wire1 down
        0x27,3, // horizontal wire2
        0x2B,3, // vertical wire1
        0x2C,3, // vertical wire3
        0x2D,3, // decoration, wants wire3 down, can take wires left
        0x2E,3, // decoration, wants wire2 right, can take wires up
        0x2F,3, // background
        0x34,3, // background shadow
-1,'W', 0x36,3, 0x37,3,
        0x3E,3, 0x3F,3, // background detail (inset)
// Walls
-1,'#', 0x60,0, 0x61,0, 0x62,0,
        0x68,0, 0x69,0, 0x6A,0, // wall
-1,'#', 0x70,0, // single block platform
        0x60,0, 0x61,0, 0x62,0, // longer platform (edges used only if not anchored in wall)
-1,'#', 0x72,0, // a decoration inside the wall, randomly appearing

-1,'A', 0x73,1, // arc generator
-1,'$', 0x7C,1,0x7D,1, // arcbox (wall decoration)
-1,'R', 0x75,0, // one type of vertical wall (or horizontal!)
-1,'R', 0x71,0, // another type of vertical wall
-1,'#', 0x51,0, 0x52,0, 0x53,0, // Wily signature
// Ladder
-1,'|', 0x83,0,
// Wily door
-1,'D', 0x10,0,
// Gutsblock
-1,'G', 0x38,0, 0x39,0,
        0x3A,0, 0x3B,0,
// no spikes
-2);
}
static bool ElecAir(char c)
{
    return c==' ' || c==',' || c=='.' || c=='W';
}
static bool ElecWall(char c)
{
    return c=='R' || c=='S' || c=='#' || c=='&' || c=='A' || c=='$';
}
static bool ElecVoid(char c)
{
    return ElecAir(c) && c != ' ';
}
static bool ElecHidden(char c)
{
    return c != ' ' && !ElecWall(c) && !ElecVoid(c);
}

namespace ElecUtil
{
    class GenWires
    {
    private:
        enum
        {
            EndPoint=1,
            Hidden  =2,
            Drawable=4,
            
            Void  = 0              | 0x08,
            Wall  = EndPoint       | 0x10,
            Ladder= Hidden|Drawable| 0x20,
            Other = Drawable       | 0x40
        };
        struct Task
        {
            int dir; // -1,+1,-16,+61
            unsigned char symbol;
            unsigned point;
        };
    public:
        GenWires(unsigned char* decorations, const char* ForegroundMap)
          : result(decorations)
        {
            for(unsigned y=0; y<16; ++y)
                for(unsigned x=0; x<16; ++x)
                {
                    unsigned a = y*16+x;
                    unsigned pos = y*32+x;
                    if(ElecWall(ForegroundMap[pos])) Classify[a] = Wall;
                    else if(ElecVoid(ForegroundMap[pos])) Classify[a] = Void;
                    else if(ElecHidden(ForegroundMap[pos])) Classify[a] = Ladder;
                    else if(y > 0 && ElecWall(ForegroundMap[pos-32])) Classify[a] = Void;
                    else Classify[a] = Other;
                }
            Gen();
        }
    private:
        void Gen()
        {
            unsigned nvisible=0;
            for(unsigned a=0; a<256; ++a)
                if(IsDrawable(a) && !IsHidden(a)) ++nvisible;
            
            unsigned nbox = 1 + (nvisible*3/256);
            for(unsigned a=0; a<nbox; ++a)
            {
                unsigned point = FindFarthestPoint();
                
                unsigned char boxtype = 0x2D + (a&1);
                
                unsigned groupsize=1 + (a&1);
                PutBoxIn(point, boxtype, groupsize);
                RunTasks();
            }
        }
        
        void RunTasks()
        {
            while(!tasks.empty())
            {
                const Task t = tasks[0];
                tasks.erase(tasks.begin());
                RunTask(t);
            }
        }
        
        void RunTask(const Task& t)
        {
            /* Note: Don't plot into the starting point. It is the intersection
             * where this wire was launched from.
             */
            
            unsigned point = t.point;
            const int dir = t.dir;
            const unsigned char symbol = t.symbol;
            
            const int xdir = dir==-1 ? -1 : dir==1 ? 1 : 0;
            const int ydir = dir/16;
            
            std::vector<unsigned> donepoints;
            
            bool cancel = false;
            
            for(;;)
            {
                unsigned x = point%16;
                unsigned y = point/16;

                if((x+xdir) < 0 || (x+xdir) > 15
                || (y+ydir) < 0 || (y+ydir) > 15)
                {
                    cancel=true;
                    break;
                }

                if(IsEndPoint(point + dir))
                {
                    // The wire was delivered successfully!
                    break;
                }
                
                if(Plot(point+dir, symbol))
                {
                    donepoints.push_back(point+dir);
                    point += dir;
                    continue;
                }
                
                // The wire couldn't be delivered successfully.
                cancel = true;
                break;
            }
            
            // The wire didn't work that way. Cancel the task.

            unsigned remain = donepoints.size();
            if(cancel && remain > 0)
            {
                unsigned char possiblebox = 0;
                switch(symbol)
                {
                    // three horiz
                    case 0x25: if(xdir > 0) possiblebox = 0x2D;
                               if(xdir < 0) possiblebox = 0x2E;
                               break;
                    // two horiz
                    case 0x27: if(xdir > 0) possiblebox = 0x2D;
                               //if(xdir < 0) possiblebox = 0x2E; - indirection failure
                               break;
                    // one vert
                    case 0x2B: if(ydir > 0) possiblebox = 0x2E;
                               break;
                    // three vert
                    case 0x2C: if(ydir > 0) possiblebox = 0x2E;
                               if(ydir < 0) possiblebox = 0x2D;
                               break;
                }
                
                unsigned boxplace = rand()%(remain+1);
                
                for(unsigned a=remain; a-->0; )
                {
                    Unplot(donepoints[a]); --remain;
                    if(a==boxplace)
                    {
                        unsigned groupsize=1;
                        PutBoxIn(donepoints[a], possiblebox, groupsize);
                        break;
                    }
                }
            }
            donepoints.resize(remain);
            
            if(donepoints.empty()) return;
            
            unsigned char possiblejunction = 0x00, possiblechange = 0x00;
            switch(xdir * symbol)
            {
                case -0x25: // right edge=3 wires
                {
                    possiblejunction = 0x24; // right=3, left=2.
                    possiblechange   = 0x25; // indirection failure. 0x27;
                    break;
                }
                case -0x27: // right edge=2 wires
                {
                    possiblejunction = 0x26; // right=2, left=3.
                    possiblechange   = 0x25;
                    break;
                }
                case +0x25: // left edge=3 wires
                {
                    possiblejunction = 0x26; // left=3, right=2.
                    possiblechange   = 0x27;
                    break;
                }
                case +0x27: // left edge=2 wires
                {
                    possiblejunction = 0; // indirection failure. 0x24; // left=2, right=3.
                    possiblechange   = 0x25;
                    break;
                }
            }
            
            int junctiondir = 0;
            switch(possiblejunction)
            {
                case 0x24:
                    junctiondir = -16;
                    break;
                case 0x26:
                    junctiondir = 16;
                    break;
            }
            if(junctiondir)
            {
                int jpoint = FindBestJunctionPoint
                    (donepoints[0], donepoints[donepoints.size()-1],
                     dir, junctiondir);
                if(jpoint >= 0) // && std::rand()%100 < 100)
                {
                    bool change=false;
                    for(unsigned a=0; a<donepoints.size(); ++a)
                    {
                        if(donepoints[a] == jpoint)
                        {
                            result[jpoint] = possiblejunction;
                            AddTask(jpoint, junctiondir, 0x2B); // 1-wire to that direction
                            change=true;
                        }
                        else if(change)
                        {
                            result[donepoints[a]] = possiblechange;
                        }
                    }
                }
            }
        }
        
        int FindBestJunctionPoint(unsigned begin,unsigned end, int dir, int jdir)
        {
            const int xdir = jdir==-1 ? -1 : jdir==1 ? 1 : 0;
            const int ydir = jdir/16;
            
            unsigned range = 1;
            for(unsigned point=begin; point!=end; point+=dir) ++range;
            
            unsigned bestpoint=0;
            unsigned bestlength=0;
            unsigned rangepos=0;
            for(unsigned point=begin; ; point+=dir)
            {
                unsigned testpoint = point;
                unsigned length = 0;
                for(;;)
                {
                    unsigned x = testpoint%16;
                    unsigned y = testpoint/16;
                    
                    if((x+xdir) < 0 || (x+xdir) > 15
                    || (y+ydir) < 0 || (y+ydir) > 15)
                    {
                        break;
                    }
                    if(IsEndPoint(testpoint+jdir) || !IsDrawable(testpoint+jdir)) break;
                    ++length;
                    testpoint += jdir;
                }
                
                length = length*10 + (rangepos*10/range);
                
                if(!(testpoint & 0x11)) // odd positions aren't nice
                {
                    if(length > bestlength) { bestlength=length; bestpoint=point; }
                }
                
                if(point == end) break;
                ++rangepos;
            }
            if(bestlength < 10) return -1;
            return bestpoint;
        }
        
        void PutBoxIn(unsigned point, unsigned char boxtype, unsigned& groupsize)
        {
            unsigned x=point%16, y=point/16;
            
            if(!Plot(point, boxtype)) return;
            --groupsize;
            
            switch(boxtype)
            {
                case 0x2D:
                    // leftside: anything goes
                    if(x>0)  AddTask(point,  -1, std::rand()%2 ? 0x25 : 0x27);
                    // downside: three
                    if(y<15) AddTask(point,  16, 0x2C);
                    
                    if(y>0  && groupsize>0) PutBoxIn(point-16, 0x2E, groupsize);
                    if(x<15 && groupsize>0)  PutBoxIn(point+1, 0x2E, groupsize);
                    break;
                case 0x2E:
                    // rightside: two (indirection failure, use 3).
                    if(x<15) AddTask(point,  +1, 0x25);
                    // topside: anything goes
                    if(y>0)  AddTask(point, -16, std::rand()%2 ? 0x2B : 0x2C);

                    if(x>0 && groupsize>0)  PutBoxIn(point-1, 0x2D, groupsize);
                    if(y<15 && groupsize>0) PutBoxIn(point+16, 0x2D, groupsize);
                    break;
            }
        }
        
        void AddTask(unsigned point, int dir, unsigned char symbol)
        {
            Task t;
            t.point  = point;
            t.dir    = dir;
            t.symbol = symbol;
            tasks.push_back(t);
        }
        
        bool Plot(unsigned point, unsigned char symbol)
        {
            if(!IsDrawable(point)) return false;
            result[point] = symbol;
            Classify[point] &= ~Drawable;
            return true;
        }
        void Unplot(unsigned point)
        {
            result[point] = 0;
            Classify[point] |= Drawable;
        }
        
        inline bool IsDrawable(unsigned pos) const { return Classify[pos] & Drawable; }
        inline bool IsEndPoint(unsigned pos) const { return Classify[pos] & EndPoint; }
        inline bool IsHidden(unsigned pos) const { return Classify[pos] & Hidden; }
        
        unsigned FindFarthestPoint() const
        {
            /* Find the farthest point from all walls / airs */
            /* The farthest point is the one that has biggest
             * minimum distance to NonDrawable. */
            int dist[16*16];
            for(unsigned pos=0; pos<256; ++pos)
            {
                dist[pos] = 0xFF;
                if(Classify[pos] == Wall) dist[pos] = 0x00;
                else if(Classify[pos] == Void) dist[pos] = 0x00;
                else if(!IsDrawable(pos)) dist[pos] = -5;
            }
            /* Fill the map with distances */
            for(unsigned i=0; i<15; ++i)
                for(unsigned y=0; y<16; ++y)
                    for(unsigned x=0; x<16; ++x)
                    {
                        int d = dist[y*16+x];
                        d = std::min(d, 1 + (y>0 ? dist[(y-1)*16+(x)] : 0));
                        d = std::min(d, 1 + (y<15? dist[(y+1)*16+(x)] : 0));
                        d = std::min(d, 1 + (x>0 ? dist[(y)*16+(x-1)] : 0));
                        d = std::min(d, 1 + (x<15? dist[(y)*16+(x+1)] : 0));
                        
                        dist[y*16+x] = d;
                    }
            int bestdist=0, best=8*16+8;
            for(unsigned pos=0; pos<256; ++pos)
            {
                int d = dist[pos];
                
                if(pos & 0x11) d -= 3; // value even positions 

                //if(d > 2) d = 2;
                if(d > bestdist) { bestdist=dist[pos]; best=pos; }
            }
            return best;
        }
    private:
        unsigned char* result;
        char Classify[16*16];

        std::vector<Task> tasks;
    };
}

void ElecmanChar(Screen& s)
{
    unsigned randseed = 0;
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
            randseed += s.ascii[y*32+x];
    std::srand(randseed);
    //std::srand(0);
    
    unsigned char decorations[16*16] = {0};
    
    unsigned char Lego1[2][3] = {{0x60,0x61,0x62},{0x68,0x69,0x6A}};
    
    LegoFill(decorations, s.ascii,'#', Lego1,Lego1);
    LegoFill(decorations, s.ascii,'&', Lego1,Lego1);

    // Find single-block-height platforms.
    for(unsigned y=1; y<15; ++y) 
        for(unsigned x=0; x<16; ++x)
        {
            if(s.ascii[(y)  *32+x]=='#'
            && ElecAir(s.ascii[(y-1)*32+x])
            && ElecAir(s.ascii[(y+1)*32+x]))
            {
                // platform that is wholly in air!
                
                // Find the left edge
                unsigned begin = x;
                while(begin > 0)
                {
                    unsigned testx = begin-1;
                    bool air = ElecAir(s.ascii[(y-1)*32+testx]);
                    bool wall=!ElecAir(s.ascii[(y  )*32+testx]);
                    if(!wall || !air) break;
                    begin=testx;
                }
                // Find the right edge
                unsigned end = x;
                while(end < 15)
                {
                    unsigned testx = end+1;
                    bool air = ElecAir(s.ascii[(y-1)*32+testx]);
                    bool wall=!ElecAir(s.ascii[(y  )*32+testx]);
                    if(!wall || !air) break;
                    end=testx;
                }
                unsigned length = end-begin+1;
				if(length == 1)
					decorations[y*16+begin] = 0x70; // single block platform
				else
				{
				    bool anchored = begin==0 || !ElecAir(s.ascii[y*32+begin-1]);
					decorations[y*16+begin] = anchored ? 0x61 : 0x62;
		        }
			    for(unsigned xpos=begin+1; xpos<end; ++xpos)
			        decorations[y*16+xpos] = 0x61; // platform
				if(length != 1)
				{
				    bool anchored = end==15 || !ElecAir(s.ascii[y*32+end+1]);
					decorations[y*16+end] = anchored ? 0x61 : 0x62;
		        }
		        x = end;
            }
        }
    
    ElecUtil::GenWires(decorations, s.ascii);
    
    // Convert to tiles
    for(unsigned y=0; y<16; ++y) 
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c   = 0;
            unsigned char pal = 0;
            
            switch(s.ascii[y*32+x])
            {
                case '.':
                {
                    // background
                    pal = 0;
                    c   = 0x01;
                    break;
                }
                case ',':
                {
                    // void
                    pal = c = 0;
                    break;
                }
                case ' ':
                {
                    // background
                    pal = 3;
                    if(y>0 && ElecWall(s.ascii[(y-1)*32+x]))
                        c = 0x34;
                    else
                    {
                        char deco = decorations[y*16+x];
                        c   = deco ? deco : 0x2F;
                    }
                    break;
                }
                case '#':
                case '&':
                {
                    char deco = decorations[y*16+x];
                    pal = s.ascii[y*32+x] == '#' ? 0 : 2;
                    c   = deco ? deco : 0x70;
                    break;
                }
                case 'A':
                    c = 0x73;
                    pal = 1;
                    break;
                case '$':
                    pal = 1;
                    c = 0x7C + (x&1);
                    break;
                case 'S':
                case 'R':
                {
                    c = s.ascii[y*32+x];
                    // air right/left?
                    bool right = x<15 && (s.ascii[y*32+(x+1)]!=c && s.ascii[y*32+(x+1)]!='|');
                    bool left  = x>0  && (s.ascii[y*32+(x-1)]!=c && s.ascii[y*32+(x-1)]!='|');
                    // this type up/down?
                    bool up   = y>0  && (s.ascii[(y-1)*32+x]==c);
                    bool down = y<15 && (s.ascii[(y+1)*32+x]==c);
                    
                    pal = c=='R' ? 0 : 2;
                    c = ((right||left)&&(up||down))
                     || ((x==0 || x==15) && up&&down)
                            ? 0x71 : 0x75;
                    // 0x71 is when there's air next to it but no companion vertically
                    break;
                }
                case Ascii_Ladder:
                case '[':
                    c   = 0x83; // ladder
                    pal = s.ascii[y*32+x]=='[' ? 2 : 0;
                    break;
                case Ascii_Spike:
                    c   = 0xC4; // spike
                    if(y>0 && s.ascii[(y-1)*32+x] == ' ')
                        pal = 0x02;
                    else
                        pal = 0x00;
                    break;
                case Ascii_WilyDoor:
                    c   = 0x10;
                    pal = 0x00;
                    break;
                case Ascii_Guts:
                    c   = 0x38 + (x&1) + (y&1)*2; // gutsblock
                    pal = 0x00;
                    break;
                case 'W':
                    c   = 0x36 + (x&1) + (y&1)*8; // window in bg
                    pal = 0x03;
                    break;
                default:
                    printf("? '%c'\n", s.ascii[y*32+x]);
            }
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = pal;
        }
}

/////////////////////////////////////////////////////////////

/* GUTSMAN
ti pa

01.1       = air

3E.0 3F.0
20.0 28.0 = background: mountain

3C.2       = background: mountain edge (top left = sky)
3D.2       = background: mountain edge (top right = sky)

37.3       = background: fence

25.0       = background: hole in mountain (bottom right = darkness)
26.0       = background: hole in mountain (bottom left = darkness)
2C.0       = background: \ crack in mountain
0F.0       = background: / crack in mountain
34.0       = background: / crack in mountain (another look)
00.0       = background: darkness

35.0       = background: inside a tunnel

70.1       = horizontal steel bar (when rock below)
73.1       = horizontal steel bar (when air below)
67.1       = bridge foot (probably wall)

6E.1       = vertical steel bar (anchored at left side)
6F.1       = vertical steel bar
76.1       = vertical steel bar (anchored at right side)

71.3 72.3
69.3 6A.3 = wall
6B.3       = wall right edge
6D.3       = wall bottom edge

21.2       = lift path edge (*) (sky background)
22.2       = lift path      (-)
23.2       = lift path hole (/)

84.1       = spike up (sky bg)



*/
void GutsmanAscii(Screen& s)
{
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = s.outmap[y*16+x].tile;
            char ch = '?';
            /* Classify types.
             *
             *
             *
             */
            switch(c)
            {
                case 0x01:
                    ch = '.'; // sky
                    break;
                case 0x71: case 0x72:
                case 0x69: case 0x6A:
                case 0x6B: case 0x6D: // rock wall.
                case 0x70: case 0x73: // horizontal steel bar.
                case 0x51: case 0x52: case 0x53: // wily signature
                    ch = '#';
                    break;
                case 0x67: // bridge foot
                    ch = 'X';
                    break;
                case 0x6E: case 0x6F:
                case 0x76: // vertical steel bar.
                    ch = '&';
                    break;
                
                case 0x3C ... 0x3F:
                case 0x20: case 0x28:
                case 0x25: case 0x26:
                case 0x2C: case 0x0F: case 0x34:
                    ch = ' '; // mountain background
                    break;
                
                case 0x00: // dark background
                case 0x37: // fence
                case 0x35: // inside background
                    ch = ',';
                    break;
                
                case 0x10:
                    ch = Ascii_WilyDoor;
                    break;
                
                case 0x84:
                    ch = Ascii_Spike;
                    break;
                
                case 0x38:
                case 0x39:
                case 0x3A:
                case 0x3B:
                    ch = Ascii_Guts;
                    break;
                
                case 0x21: ch = '*'; break;
                case 0x22: ch = '-'; break;
                case 0x23: ch = '/'; break;

                // no ladders
            }
            
            s.ascii[y*32+x] = ch;
            s.ascii[y*32+x+16] = ' ';
            //printf("%02X",c);
        }
}

void GutsmanChar(Screen& s) { DummyChar(s); }

/////////////////////////////////////////////////////////////

void Wily1Ascii(Screen& s)
{
    GenericAsciiMaker(s,
// Boss room
-1,'X', 0x00,1,
// Suspense
-1,'V', 0x00,0, // entire black background
// Sky
-1,'.', 0x01,1,
// Decorations
-1,' ', 0x16,2, 0x17,2, // bg, shadow
-1,' ', 0x27,1, // grass
// Walls
-1,'#', 0x48,3, 0x49,3, 0x4A,3,
        0x50,3, 0x51,3, 0x52,3, // block type 1 (middle piece repeatable)
-1,'R', 0x4B,0, 0x4C,0, 0x4D,0,
        0x53,0, 0x54,0, 0x55,0, // block type 2 (middle piece repeatable)
-1,'#', 0x66,0, // floor outdoors
-1,'#', 0x4E,3, // floor indoors
// Spikes up, down (black bg)
-1,'^', 0xAE,0, 
-1,'^', 0xBF,0,
// Ladder
-1,'|', 0xC5,0,
// Gutsblock
-1,'G', 0x38,0, 0x39,0,
        0x3A,0, 0x3B,0,
-2);
}

static bool Wily1Air(char c)
{
    return c==' ' || c=='V';
}
static bool Wily1Wall(char c)
{
    return c=='#' || c=='R';
}

void Wily1Char(Screen& s)
{
    std::srand(0);
    
    unsigned char decorations[16*16] = {0};
    
    unsigned char Lego1[2][3] = {{0x48,0x49,0x4A},{0x50,0x51,0x52}};
    unsigned char Lego2[2][3] = {{0x4B,0x4C,0x4D},{0x53,0x54,0x55}};
    
    LegoFill(decorations, s.ascii,'#', Lego1,Lego1);
    LegoFill(decorations, s.ascii,'R', Lego2,Lego2);

    // Find single-block-height platforms.
    for(unsigned y=1; y<15; ++y) 
        for(unsigned x=0; x<16; ++x)
        {
            if(s.ascii[(y)*32+x]=='#'
            && Wily1Air(s.ascii[(y-1)*32+x])
            && Wily1Air(s.ascii[(y+1)*32+x]))
            {
                // platform that is wholly in air!
                
                // Find the left edge
                unsigned begin = x;
                while(begin > 0)
                {
                    unsigned testx = begin-1;
                    bool air = Wily1Air(s.ascii[(y-1)*32+testx]);
                    bool wall=!Wily1Air(s.ascii[(y  )*32+testx]);
                    if(!wall || !air) break;
                    begin=testx;
                }
                // Find the right edge
                unsigned end = x;
                while(end < 15)
                {
                    unsigned testx = end+1;
                    bool air = Wily1Air(s.ascii[(y-1)*32+testx]);
                    bool wall=!Wily1Air(s.ascii[(y  )*32+testx]);
                    if(!wall || !air) break;
                    end=testx;
                }
                for(unsigned xpos=begin; xpos<=end; ++xpos)
                    decorations[y*16+xpos] = 0x4E; // platform
		        x = end;
            }
        }

    // Convert to tiles
    for(unsigned y=0; y<16; ++y)
        for(unsigned x=0; x<16; ++x)
        {
            unsigned char c = 0;
            unsigned char p = 0;
            
            /* Wily1 palettes:
             *   
             * 0 = solid
             * 1 = blacks
             * 2 = blues
             * 3 = blue,gray
             */
            
            switch(s.ascii[y*32+x])
            {
                case 'X': // boss room bg
                    c   = 0x00;
                    p   = 1;
                    break;
                case 'V': // black bg
                    c   = 0x00;
                    p   = 0;
                    break;
                case '.': // sky
                    c   = 0x01;
                    p   = 1;
                    break;
                case ' ':
                {
                    // If above is sky, this is grass.
                    if(y>0 && s.ascii[(y-1)*32 + x]=='.')
                    {
                        c = 0x27;
                        p = 1;
                        break;
                    }
                    c = 0x17;
                    p = 2;
                    // If above is solid, this is shadow.
                    if(y>0 && Wily1Wall(s.ascii[(y-1)*32 + x]))
                    {
                        c = 0x16;
                    }
                    break;
                }
                case Ascii_Spike:
                {
                    // If below is solid, it's up. Otherwise it's down.
                    if(y<16 && Wily1Wall(s.ascii[(y+1)*32 + x]))
                        c = 0xAE;
                    else
                        c = 0xBF;
                    p = 0;                    
                    break;
                }
                case Ascii_Ladder:
                {
                    c = 0xC5;
                    p = 0;
                    break;
                }
                case Ascii_Guts:
                {
                    c = 0x38 + (x&1) + (y&1)*2; // gutsblock
                    p = 0;
                    break;
                }
                case '#':
                case 'R':
                {
                    char deco = decorations[y*16+x];
                    p = 3;
                    c = 0x66; if(deco) c = deco;
                    if((c >= 0x4B && c <= 0x4D) || (c >= 0x53 && c <= 0x55)) p = 0;
                    break;
                }
                default:
                    printf("? '%c'\n", s.ascii[y*32+x]);
            }
            
            s.outmap[y*16+x].tile    = c;
            s.outmap[y*16+x].palette = p;
        }
}

/////////////////////////////////////////////////////////////

void Wily2Ascii(Screen& s)
{
    GenericAsciiMaker(s,
// Sky
-1,'.', 0x01,2,
// Shadow
-1,',', 0x00,0,
// Walls
-1,'#', 0x48,1, 0x49,1, 0x4A,1,
        0x50,1, 0x51,1, 0x52,1, // block type 1 (middle piece repeatable)
-1,'&', 0x0A,1, 0x09,1, 0x0A,1,
        0x10,1, 0x11,1, 0x12,1, // pseudo block (looks same as block 1)
-1,'#', 0x4E,1, // platform
-1,'#', 0x61,2, // / edge (bottomside)
-1,'#', 0x60,2, // \ edge (bottomside)
-1,'#', 0x62,2, // / edge (topside)
// Decorations
-1,' ', 0x17,3, 0x16,3, // inside bg, inside shadow
-1,' ', 0x23,2, 0x24,2, // window (preferably 2x2)
// Spike
-1,'^', 0xC4,2,
// Ladder
-1,'|', 0x85,0,
// Boss bg (random colours)
-1,'X', 0x25,3, 0x26,3, 0x27,3,
-2);
}

void Wily2Char(Screen& s) { DummyChar(s); }


/////////////////////////////////////////////////////////////

void Wily3Ascii(Screen& s)
{
    GenericAsciiMaker(s,
// void
-1,'.', 0x00,0,
// ground
-1,'#', 0x60,3, // ground
        0x61,0, // horizontal edge
        0x69,0, // junction
        0x68,0, // vertical edge
-1,'-', 0x2D,1, // horiz shadow
        0x2E,1, // horiz bg
        0x2F,1, // horiz bottom
-1,'=', 0x2D,2,
        0x2E,2,
        0x2F,2, // same in blue
-1,'|', 0x22,1, // vertical topleft shadow
        0x23,1, // vertical top shadow
        0x24,1, // vertical topright shadow
        0x2A,1, // vertical left edge
        0x2B,1, // vertical bg
        0x2C,1, // vertical right edge
-1,'!', 0x22,2,
        0x23,2,
        0x24,2,
        0x2A,2,
        0x2B,2,
        0x2C,2, // same in blue
-1,'G', 0x38,0,0x39,0,
        0x3A,0,0x3B,0, // gutsblock
-2);
}

void Wily3Char(Screen& s) { DummyChar(s); }

/////////////////////////////////////////////////////////////

void Wily4Ascii(Screen& s)
{
GenericAsciiMaker(s,
// Walls
-1,'#', 0x48,3, 0x49,3, 0x4A,3,
        0x50,3, 0x51,3, 0x52,3, // block type 1 (middle piece repeatable)
-1,'#', 0x4B,0, 0x4C,0, 0x4D,0,
        0x53,0, 0x54,0, 0x55,0, // block type 2 (middle piece repeatable)
-1,'#', 0x6F,0, // vertical block
-1,'#', 0x4E,3, // inside floor
-1,'#', 0x64,1, 0x65,1,
        0x6C,1, 0x6D,1, // gutsman decoration
// Decorations
-1,' ', 0x17,2, 0x16,3, // inside bg, inside shadow
// Ladder
-1,'|', 0xC5,0,
// Lift
-1,'*', 0x3E,0, // edge
-1,'-', 0x3C,0, // path
-1,'/', 0x3D,0, // hole
// Spike
-1,'^', 0xAE,0, // black bg
// Various backgrounds
-1,'.', 0x00,0, // black background
-1,'1', 0x00,1, // palette 1 (wily room)
-1,'2', 0x00,2, // palette 2 (wily room)
// Teleport
-1,'_', 0x63,1, 0x69,1, // top
-1,'=', 0x68,1, 0x6A,1, // bottom
// Wily room door
-1,'D', 0x06,0,
// Wily room floor
-1,'W', 0x62,0,
-2);
}

void Wily4Char(Screen& s) { DummyChar(s); }

/////////////////////////////////////////////////////////////

void GameEndAscii(Screen& s)
{
GenericAsciiMaker(s,
// Void
-1,' ', 0x01,0, // void
-1,'.', 0x00,0, // sky
// Ground
-1,'#', 0x70,1,
// Field, also dock
-1,'-', 0x3C,2,
// Forests
-1,'f', 0x32,1, // with mountain background
-1,'f', 0x31,1, // without mountain background
// Mountains
-1,'/', 0x20,3, 0x28,3,
-1,'L', 0x21,3,
-1,'M', 0x0C,3, 0x07,3, 0x29,3, 0x26,3, 0x27,3, 0x23,3, 0x27,3, 0x25,3, 0x24,3, 0x22,3,
// Sea
-1,'(', 0x3D,2, // sea edge
-1,')', 0x3F,2, // sea edge
-1,'~', 0x3E,2, // ground-sea edge
        0x37,2, // just sea
-1,'B', 0x34,2, 0x35,2, // boat on dock (3x1)
-1,'=', 0x36,2, // dock farther
// Buildings
-1,'|', 0x09,3, 0x33,3, 0x08,3, // of different height and composition
        0x0A,3, 0x3A,3, 0x33,3, // 0x08 is a /\ top of a building
                        0x38,3, // 38,39,3A,3B can be stacked randomly.
                        0x39,3,
                        0x3A,3,
                        0x3B,3,
// Roll's house
-1,'H', 0x2A,0,
        0x2E,0, 0x2F,0, // roof (2E... 2F)
        0x0D,0, // right edge
-1,'P', 0x0E,0, // right edge with a smoke pipe
-1,'W', 0x2B,0, // window in Roll's house

-2);
}

void GameEndChar(Screen& s) { DummyChar(s); }
