#define OBJECTS_CC
#include "defs.hh"

#include "objects.hh"

// reserved: X, S, I, Y, 0-9

const struct ObjectDesc AsciiObjects[] =
{
  //  0
  { 'y', 0x0,0xC }, // bunby heri/blader - the annoying thing in Cutman level
                    // [1] 6E (palette 1)
  { 'm', 0x0,0x8 }, // metto
                    // [5] 6A (palette 1)
  { 'j', 0x8,0x8 }, // red jumping junk (Flea)
                    // [2] 7E (palette 1)
  { 'f', 0x8,0xC }, // floor trap/landmine (Spine/Gabyoall)
                    // [5] 72 (palette 2)
  //  4
  { 'c', 0x8,0x8 }, // vertically moving eyecube (suzy)
                    // [1] 6E (palette 2)
  { 'v', 0x8,0x0 }, // spread cannon in floor
                    // [4-5] 7A,7E (palette 2)
  { 0 },   // SPECIAL, SEE BELOW
  { 'i', 0x0,0x0 }, // sinewave flying penguin
                    // [4] 76 (blackwhite-palette)
  //  8
  { 's', 0xF,0x0 }, // sinewave flying bomb
                    // [4] 62 (palette 2)
  { 't', 0x0,0xC }, // jumping telephone post (Big Eye) (bomb damage only 4)
                    // [2-3] C2,C6 (palette 1)
  { 'u', 0x8,0x0 }, // hamburger
                    // [5] 72 (palette 2)
  { '<', 0x8,0x8 }, // singleshot cannon pointing left (timing 1)
                    // [4] 62 (palette 2)
  //  C
  { '-', 0x0,0x8 }, // elec arc
                    // [4] 66 (palette 1)
  { 'B', 0x8,0x0 }, // periodical bombing (y = 0xE0) (not destroyable)
                    // [1] BA (palette 2)
  { 0 },   // Zig-Zag fire. (SPECIAL, SEE BELOW)
  { 'I', 0x8,0x0 }, // flying fire, take two! (not destroyable)
                    // [1] 7A (palette 1) (appears behind walls!)
  // 10
  { 0 },   // Buzzblock engine. (SPECIAL, SEE BELOW)
  { 0 },   // UNKNOWN (could be Scissor machine's scissor)
  { 'G', 0x0,0x4 }, // green guardian (Xl=0, Yl=4)
                    // [1-3] 96,9A,9E (palette 1)
  { 'L', 0x0,0xC }, // shooting lift ("foot holder") (not destroyable)
                    // [2-3] B6,BA (palette 2)
  // 14
  { 'R', 0x0,0x4 }, // robots that may lose their head
                    // [1-3] A2,A6,AA (palette 1)
  { 'a', 0x0,0x4 }, // axe throwing robot in Gutsman
                    // [2-3] 82,86 (POSSIBLE)
  { '.', 0x8,0x0 }, // spread cannon pointing down
                    // [4-5] 7A,7E (palette 2)
  { '>', 0x8,0x8 }, // singleshot cannon pointing right
                    // [4] 62 (palette 2)
  // 18
  { ']', 0x4,0x6 }, // fire torch to the right (Fireman)
                    // [4] C6 (palette 1) (incomplete?)
  { '', 0x0,0x7 }, // electricity decoration 1 (Elecman level only)
                    // [4] 66 (palette blackwhite)
  { 0 }, // bullet sound? (not destroyable)
  { 0 }, //small explosion (from object $0D but also from enemy killed) (not destroyable)
  // 1C
  { 'r', 0x0,0x4 }, // scissor machine
  { 0 }, //instantiated at $B222, $BC73
  { 0 }, //instantiated at $BD1F
  { 0 }, //instantiated at $B752
  // 20
  { 0 }, // UNKNOWN
  { 'h', 0x8,0x8 }, // horizontally moving eyecube (suzy)
                    // [1] 6E (palette 2)
  { '{', 0x8,0x8 }, // singleshot cannon pointing left (timing 2)
                    // [4] 62 (palette 2)
  { '}', 0x8,0x8 }, // singleshot cannon pointing up
                    // [4] 62 (palette 2)
  // 24
  { '=', 0x0,0x8 }, // elec arc, different timing
                    // [4] 66 (palette blackwhite)
  { '', 0x0,0x7 }, // electricity decoration 2 (Elecman level only)
                    // [4] 66 (palette blackwhite)
  { 0 },
  { 0 }, // buzzblocks (Elecman stage)
  // 28
  { 0 }, //instantiated at $AF19
  { 0 }, //instantiated at $B218, $B762, also referenced at C6C6 (I think)
  { 0 }, //instantiated at $AD6A
  { 0 },
  // 2C
  { 0 }, // lift engine (SPECIAL)
         // [2-3] AE,B2
  { 0 }, //bullet sound? instantiated at $E7F8
  { 0 },
  { '',0,0 },  // lift SFX (Gutsman) or Wily3 tunnel manager (SPECIAL)
  // 30
  { '',0,0 },  // stop SFX (Gutsman) or setup BG pal (wily3) (SPECIAL)
  { 'd', 0x0,0x0 }, // periodically three fire devils from lava to sky
  { '', 0,0 },   // rotates palettes (SPECIAL)
  { 0 },  // scissor weapon
  // 34
  { 0 },  // ice weapon
  { 0 },  // bomb weapon
  { 0 },  // fire weapon
  { 0 },  // fire shield weapon
  // 38
  { 0 }, // Boss6's pieces
  { 0 },
  { 0 },   // wily3 boss
  { 'T', 0x4,0xC }, // teleport when a "boss" killed, activate enemy if not dead.
  // 3C
  { 'o', 0x8,0x9 }, // score ball = bonus pearl
                    // (palette 2)
  { 'p', 0x8,0xC }, // small weapon
  { 'e', 0x8,0x9 }, // small energy
  { 'P', 0x8,0x8 }, // big weapon
  // 40 
  { 'E', 0x8,0x8 }, // big energy
  { 'U', 0x8,0x8 }, // 1up
  { 'M', 0x8,0x8 }, // magnet beam
                    // [3] 96 (palette Megaman)
  { 'O', 0x8,0x0 }, // teleport immediately (invisible)
  // 44
  { '*', 0x8,0x8 }, // end level
                    // (palette 1)
  { 'W', 0x8,0x8 }, // all weapons full (yashichi)
                    // [5] 72 (palette 1)
  { 0 },
  { 0 },
  // 48
  { 0 },
  { '', 0x8,0x8 }, // random powerup (enemy destruction)
  { 0 }, //? used at 0001CB5B
  
  /*
            [1] 8A = unknown
            
            [2] 92,96 = watchers (92=watcher, 96=electricity) (palette 1)
  
  */
};

/*
        When the Y coordinate is $FF (handled at $DA64):
            Object type $06:
                This room has "watchers" (The flying beamers in Elecman and Wily4)
            Object type $0E:
                This room has a ZigZag fire (Fireman stage)
            Object type $2C:
                This stripe contains a lift engine

        When the Y coordinate is $FE (handled at $DA32):
            Object type $09:
                ObjectFireDelay <- #$40 (and the following)
            Object type $1F:
                ObjectYSpeed <- 4, ObjectUnknown660 <- 0

        When the object type $FF (handled at $C666):
            Does patterntable writes. It is apparently used for
            changing enemy graphics during a stripe without
            requiring a scrolling to take place.
            The Y position is used as an index to table at $C6B0.

    Object 0E:
        If Y coordinate isn't $FF, it's just a normal falling fire.
    Object 10:
        A buzzblock/temporary block engine.
        The actual data this engine uses is actually hardcoded
        into the game (see defs.hh for details).
    Object 2C:
        Object 2C is a lift engine, but it mustn't be invoked as-is.
        Instead, it must be given the Y coordinate $FF

    Object 2F () (handled at 05BA6E):
        If stage = Gutsman:
          Issues the Lift sound (sfx $23)
        else
          Controls the water pipe in Wily3
    Object 30 () (handled at 05BAD0)
        If stage = Gutsman:
          Stops SFX (sfx command $FE)
        Else,
          Sets up three bytes of BG palette from 05BAF6
    Object 32:
       Rotates the lava palettes in Fireman stage.
       (The palettes (5 bytes) are at $F3F2+0).
       Also rotates the palettes in the beginning of Wily2 battle.
       (The palettes (5 bytes) are at $F3F2+5).
       
 */

const unsigned NumAsciiObjects = sizeof(AsciiObjects) / sizeof(*AsciiObjects);
